@@ -4,8 +4,8 @@
 
 ;; Author: Adam Porter <adam@alphapapa.net>
 ;; URL: http://github.com/alphapapa/magit-todos
-;; Version: 1.5.3
-;; Package-Requires: ((emacs "25.2") (async "1.9.2") (dash "2.13.0") (f "0.17.2") (hl-todo "1.9.0") (magit "2.13.0") (pcre2el "1.8") (s "1.12.0"))
+;; Version: 1.6
+;; Package-Requires: ((emacs "26.1") (async "1.9.2") (dash "2.13.0") (f "0.17.2") (hl-todo "1.9.0") (magit "2.13.0") (pcre2el "1.8") (s "1.12.0") (transient "0.2.0"))
 ;; Keywords: magit, vc
 
 ;;; Commentary:
@@ -64,7 +64,6 @@
 ;;;; Requirements
 
 (require 'cl-lib)
-(require 'org)
 (require 'grep)
 (require 'seq)
 
@@ -73,6 +72,7 @@
 (require 'f)
 (require 'hl-todo)
 (require 'magit)
+(require 'transient)
 (require 'pcre2el)
 (require 's)
 
@@ -117,11 +117,6 @@ This should be set automatically by customizing
 Used to avoid running multiple simultaneous scans for a
 magit-status buffer.")
 
-;; FIXME: Jumping binds.  In `magit-status-mode-map' now, "j" is bound
-;; to `magit-status-jump', which is a Transient command; it's no
-;; longer bound to a prefix map.  There doesn't seem to be a way to
-;; add binds to that Transient command's body.
-
 (defvar magit-todos-section-map
   (let ((map (make-sparse-keymap)))
     (define-key map "jT" #'magit-todos-jump-to-todos)
@@ -161,6 +156,8 @@ A time value as returned by `current-time'.")
 (defvar magit-todos-section-heading "TODOs"
   "Allows overriding of section heading.")
 
+(defvar org-odd-levels-only)
+
 ;;;; Customization
 
 (defgroup magit-todos nil
@@ -253,7 +250,9 @@ which the keys are the keywords.
 When set, sets `magit-todos-search-regexp' to the appropriate
 regular expression."
   :type '(choice (repeat :tag "Custom list" string)
-                 (const :tag "Keywords from `hl-todo'" hl-todo-keyword-faces)
+                 (const :tag "Keywords from `hl-todo'"
+                        :doc "Note that the keywords in `hl-todo-keyword-faces' are treated by it as regexps, while this package treats them as strings.  If this doesn't meet your needs, please use another option.  See <https://github.com/alphapapa/magit-todos/issues/101>."
+                        hl-todo-keyword-faces)
                  (variable :tag "List variable"))
   :set (lambda (option value)
          (set-default option value)
@@ -277,7 +276,9 @@ regular expression."
                  (const :tag "Never" never)))
 
 (defcustom magit-todos-buffer-item-factor 10
-  "Multiply `magit-todos-auto-group-items' and `magit-todos-max-items' by this factor in dedicated `magit-todos' buffers."
+  "Adjustment to item grouping in dedicated `magit-todos' buffers.
+Multiplies `magit-todos-auto-group-items' and
+`magit-todos-max-items' by this factor."
   :type 'integer)
 
 (defcustom magit-todos-group-by '(magit-todos-item-keyword magit-todos-item-filename)
@@ -313,6 +314,27 @@ this in a directory-local variable for certain projects."
                  (const :tag "Repo root directory only" 0)
                  (integer :tag "N levels below the repo root")))
 
+(defcustom magit-todos-insert-after '(bottom)
+  "Where to insert the TODOs section in the Magit status buffer.
+The TODOs section is inserted after the first of these sections
+found, or at the bottom if none exist.  Specific sections may be
+chosen, using the first symbol returned by evaluating
+\"(magit-section-ident (magit-current-section))\" in the status
+buffer with point on the desired section, e.g. `recent' for the
+\"Recent commits\" section.  Note that this may not work exactly
+as desired when the built-in scanner is used."
+  :type '(repeat
+          (choice (const :tag "Top" top)
+                  (const :tag "Bottom" bottom)
+                  (const :tag "Recent commits" unpushed)
+                  (const :tag "Untracked files" untracked)
+                  (const :tag "Unstaged files" unstaged)
+                  (const :tag "Staged files" staged)
+                  (const :tag "Stashes" stashes)
+                  (const :tag "Pull requests (Forge)" pullreqs)
+                  (const :tag "Issues (Forge)" issues)
+                  (symbol :tag "Specified section"))))
+
 (defcustom magit-todos-insert-at 'bottom
   "Insert the to-dos section after this section in the Magit status buffer.
 Specific sections may be chosen, using the first symbol returned
@@ -325,9 +347,17 @@ used."
                  (const :tag "Bottom" bottom)
                  (const :tag "After untracked files" untracked)
                  (const :tag "After unstaged files" unstaged)
-                 (symbol :tag "After selected section")))
+                 (symbol :tag "After selected section"))
+  :set (lambda (option value)
+         ;; For convenience, we set the new option with the appropriate value (but,
+         ;; of course, this won't work for users who set it directly with `setq'.)
+         ;; TODO: Remove this option in 1.8.
+         (ignore option)
+         (custom-set-variables `(magit-todos-insert-after ',(list value)
+                                                          'now nil "Changed by setter of obsolete option `magit-todos-insert-at'"))))
+(make-obsolete-variable 'magit-todos-insert-at 'magit-todos-insert-after "1.6")
 
-(defcustom magit-todos-exclude-globs nil
+(defcustom magit-todos-exclude-globs '(".git/")
   "Glob patterns to exclude from searches."
   :type '(repeat string))
 
@@ -359,6 +389,16 @@ from the \"topic2\" branch, this option could be set to
 \"topic\"."
   :type 'string)
 
+(defcustom magit-todos-submodule-list nil
+  "Show submodule to-do list."
+  :type 'boolean)
+
+(defcustom magit-todos-filename-filter nil
+  "Filter applied to filenames."
+  :type '(choice (const :tag "None (show filename relative to repo)" nil)
+                 (function-item :tag "Basename" file-name-nondirectory)
+                 (function :tag "Custom function")))
+
 ;;;; Commands
 
 ;;;###autoload
@@ -369,18 +409,16 @@ from the \"topic2\" branch, this option could be set to
   :global t
   (if magit-todos-mode
       (progn
-        (pcase (lookup-key magit-status-mode-map "jT")
-          ('nil (define-key magit-status-mode-map "jT" #'magit-todos-jump-to-todos))
-          ('magit-todos-jump-to-todos nil)
-          (_ (message "magit-todos: Not overriding bind of \"jT\" in `magit-status-mode-map'.")))
+        (transient-append-suffix #'magit-status-jump
+          '(0 -1) '[("T" "Todos" magit-todos-jump-to-todos)
+                    ("l" "List todos" magit-todos-list)])
         (magit-add-section-hook 'magit-status-sections-hook
                                 #'magit-todos--insert-todos
                                 nil
                                 'append)
         (add-hook 'magit-status-mode-hook #'magit-todos--add-to-status-buffer-kill-hook 'append))
     ;; Disable mode
-    (when (equal (lookup-key magit-status-mode-map "jT") #'magit-todos-jump-to-todos)
-      (define-key magit-status-mode-map "jT" nil))
+    (transient-remove-suffix #'magit-status-jump '(0 -1))
     (remove-hook 'magit-status-sections-hook #'magit-todos--insert-todos)
     (remove-hook 'magit-status-mode-hook #'magit-todos--add-to-status-buffer-kill-hook)))
 
@@ -410,6 +448,7 @@ Only necessary when option `magit-todos-update' is nil."
   (setq-local magit-todos-branch-list-merge-base-ref ref)
   (magit-todos-update))
 
+(declare-function org-show-entry "org")
 (cl-defun magit-todos-jump-to-item (&key peek (item (oref (magit-current-section) value)))
   "Show current item.
 If PEEK is non-nil, keep focus in status buffer window."
@@ -468,7 +507,11 @@ With prefix, prompt for repository."
 ;;;###autoload
 (defun magit-todos-list-internal (directory)
   "Open buffer showing to-do list of repository at DIRECTORY."
-  (magit--tramp-asserts directory)
+  (if (fboundp 'magit--tramp-asserts)
+      (magit--tramp-asserts directory)
+    (when (file-remote-p directory)
+      (magit-git-version-assert)))
+
   (let ((default-directory directory))
     (magit-setup-buffer #'magit-todos-list-mode)))
 
@@ -493,6 +536,29 @@ Type \\[magit-diff-show-or-scroll-up] to peek at the item at point."
 
 ;;;; Functions
 
+(defun magit-todos--section-end (condition)
+  "Return end position of section matching CONDITION, or nil.
+CONDITION may be one accepted by `magit-section-match', or `top'
+or `bottom', which are handled specially."
+  (cl-labels ((find-section
+               (condition) (save-excursion
+                             (goto-char (point-min))
+                             (ignore-errors
+                               (cl-loop until (magit-section-match condition)
+                                        do (magit-section-forward)
+                                        finally return (magit-current-section))))))
+    (save-excursion
+      (goto-char (point-min))
+      (pcase condition
+        ('top (when-let ((section (or (find-section 'tags)
+                                      (find-section 'tag)
+                                      (find-section 'branch))))
+                ;; Add 1 to leave blank line after top sections.
+                (1+ (oref section end))))
+        ('bottom (oref (-last-item (oref magit-root-section children)) end))
+        (_ (when-let ((section (find-section condition)))
+             (oref section end)))))))
+
 (defun magit-todos--coalesce-groups (groups)
   "Return GROUPS, coalescing any groups with `equal' keys.
 GROUPS should be an alist.  Assumes that each group contains
@@ -536,7 +602,8 @@ Chooses automatically in order defined in `magit-todos-scanners'."
 
 (cl-defun magit-todos--scan-callback (&key callback magit-status-buffer results-regexp process &allow-other-keys)
   "Call CALLBACK with arguments MAGIT-STATUS-BUFFER and match items.
-Match items are a list of `magit-todos-item' found in PROCESS's buffer for RESULTS-REGEXP."
+Match items are a list of `magit-todos-item' found in PROCESS's
+buffer for RESULTS-REGEXP."
   (funcall callback magit-status-buffer
            (with-current-buffer (process-buffer process)
              (magit-todos--buffer-items results-regexp))))
@@ -554,7 +621,8 @@ Match items are a list of `magit-todos-item' found in PROCESS's buffer for RESUL
                       ;; FIXME: This may raise multiple warnings per file.
                       (error (if (string= "Stack overflow in regexp matcher" (error-message-string err))
                                  (let ((filename (buffer-substring (point) (1- (re-search-forward ":")))))
-                                   (display-warning 'magit-todos (concat "File has lines too long for Emacs to search.  Consider excluding it from scans: " filename)))
+                                   (display-warning 'magit-todos (concat "File has lines too long for Emacs to search.  Consider excluding it from scans: " filename))
+                                   nil)
                                (signal (car err) (cdr err)))))
           (push it items))
         (forward-line 1)))
@@ -704,13 +772,6 @@ If BRANCH-P is non-nil, do not update `magit-todos-item-cache',
     (message "`magit-todos--insert-items-callback': Callback called for deleted buffer"))
   (let* ((items (magit-todos--sort items))
          (num-items (length items))
-         (group-fns (pcase magit-todos-auto-group-items
-                      ('never nil)
-                      ('always magit-todos-group-by)
-                      ((pred integerp) (when (> num-items magit-todos-auto-group-items)
-                                         magit-todos-group-by))
-                      (_ (error "Invalid value for magit-todos-auto-group-items"))))
-         (magit-todos-show-filenames (not (member 'magit-todos-item-filename group-fns)))
          (magit-section-show-child-count t)
          ;; HACK: "For internal use only."  But this makes collapsing the new section work!
          (magit-insert-section--parent magit-root-section)
@@ -732,17 +793,22 @@ If BRANCH-P is non-nil, do not update `magit-todos-item-cache',
           ;; Insert items
           (goto-char (point-min))
           ;; Go to insertion position
-          (pcase magit-todos-insert-at
-            ('top (cl-loop for ((this-section . _) . _) = (magit-section-ident (magit-current-section))
-                           until (not (member this-section '(branch tags)))
-                           do (magit-section-forward)))
-            ('bottom (goto-char (oref (-last-item (oref magit-root-section children)) end)))
-            (_ (magit-todos--skip-section (vector '* magit-todos-insert-at))))
+          (goto-char (or (cl-loop for section in magit-todos-insert-after
+                                  for pos = (magit-todos--section-end section)
+                                  when pos return pos)
+                         (magit-todos--section-end 'bottom)))
           ;; Insert section
-          (let ((reminder (if magit-todos-update
-                              "" ; Automatic updates: no reminder
-                            ;; Manual updates: remind user
-                            " (update manually)")))
+          (let* ((group-fns (pcase magit-todos-auto-group-items
+                              ('never nil)
+                              ('always magit-todos-group-by)
+                              ((pred integerp) (when (> num-items magit-todos-auto-group-items)
+                                                 magit-todos-group-by))
+                              (_ (error "Invalid value for magit-todos-auto-group-items"))))
+                 (magit-todos-show-filenames (not (member 'magit-todos-item-filename group-fns)))
+                 (reminder (if magit-todos-update
+                               ""      ; Automatic updates: no reminder
+                             ;; Manual updates: remind user
+                             " (update manually)")))
             (if (not items)
                 (unless magit-todos-update
                   ;; Manual updates: Insert section to remind user
@@ -757,7 +823,6 @@ If BRANCH-P is non-nil, do not update `magit-todos-item-cache',
                                :group-fns group-fns
                                :items items
                                :depth 0)))
-                (insert "\n")
                 (magit-todos--set-visibility :section section :num-items num-items)))))))))
 
 (cl-defun magit-todos--insert-groups (&key depth group-fns heading type items)
@@ -819,7 +884,11 @@ sections."
                                                   (if (= 1 (length group-fns))
                                                       ":" ; Let Magit add the count.
                                                     ;; Add count ourselves.
-                                                    (concat " " (format "(%s)" (length items))))))))))
+                                                    (concat " " (format "(%s)" (length items)))))))
+                          (when (= 0 depth)
+                            ;; Insert a blank line only in the body of the top-level section, so it
+                            ;; will appear only when the section is expanded, matching other sections.
+                            (insert "\n")))))
           (magit-todos--set-visibility :depth depth :num-items (length items) :section section)
           ;; Add top-level section to root section's children
           (when (= 0 depth)
@@ -853,6 +922,8 @@ sections."
                       (let* ((filename (propertize (magit-todos-item-filename item) 'face 'magit-filename))
                              (string (--> (concat indent
                                                   (when magit-todos-show-filenames
+                                                    (when magit-todos-filename-filter
+                                                      (setf filename (funcall magit-todos-filename-filter filename)))
                                                     (concat filename " "))
                                                   (funcall (if (s-suffix? ".org" filename)
                                                                #'magit-todos--format-org
@@ -888,20 +959,6 @@ automatically hidden halves at each deeper level."
          ;; Not hidden: show section manually (necessary for some reason)
          (magit-section-show section)))))
 
-(defun magit-todos--skip-section (condition)
-  "Move past the section matching CONDITION.
-See `magit-section-match'."
-  (goto-char (point-min))
-  (cl-loop until (magit-section-match condition)
-           do (magit-section-forward))
-  (cl-loop until (not (magit-section-match condition))
-           do (condition-case nil
-                  ;; `magit-section-forward' raises an error when there are no more sections.
-                  (magit-section-forward)
-                (error (progn
-                         (goto-char (1- (point-max)))
-                         (cl-return))))))
-
 (cl-defun magit-todos--line-item (regexp &optional filename)
   "Return item on current line, parsing current buffer with REGEXP.
 FILENAME is added to the item as its filename.  Sets match data.
@@ -1042,6 +1099,7 @@ if the process's buffer has already been deleted."
 
 (defun magit-todos--format-org (item)
   "Return ITEM formatted as from an Org file."
+  (require 'org)
   (magit-todos--fontify-like-in-org-mode
    (concat (magit-todos-item-org-level item) " "
            (magit-todos-item-keyword item) " "
@@ -1166,15 +1224,18 @@ It also adds the scanner to the customization variable
        ;; FIXME: That is confusing.
 
        (cl-defun ,scan-fn-symbol (&key magit-status-buffer directory depth heading sync callback)
-         ,(format "Scan for to-dos with %s, then call `%s'.
-MAGIT-STATUS-BUFFER is what it says.  DIRECTORY is the directory in which to run the scan.  DEPTH should be an integer, typically the value of `magit-todos-depth'.  HEADING is passed to `%s'.
+         ,(format "Scan for to-dos with %s.
+Then calls CALLBACK.  MAGIT-STATUS-BUFFER is what it says.  DIRECTORY
+is the directory in which to run the scan.  DEPTH should be an
+integer, typically the value of `magit-todos-depth'.  HEADING is
+passed to CALLBACK.
 
 When SYNC is nil, the scanner process is returned, and CALLBACK
 is a function which is called by the process sentinel with one
 argument, a list of match items.
 
 When SYNC is non-nil, match items are returned."
-                  name callback callback)
+                  name-without-spaces)
          (let* ((process-connection-type 'pipe)
                 (directory (f-relative directory default-directory))
                 (extra-args (when ,extra-args-var
@@ -1272,6 +1333,9 @@ When SYNC is non-nil, match items are returned."
                  (when magit-todos-exclude-globs
                    (--map (list "--glob" (concat "!" it))
                           magit-todos-exclude-globs))
+                 (unless magit-todos-submodule-list
+                   (--map (list "--glob" (concat "!" it))
+                          (magit-list-module-paths)))
                  extra-args search-regexp-pcre directory))
 
 (magit-todos-defscanner "git grep"
@@ -1287,7 +1351,10 @@ When SYNC is non-nil, match items are returned."
                  extra-args "--" directory
                  (when magit-todos-exclude-globs
                    (--map (concat ":!" it)
-                          magit-todos-exclude-globs))))
+                          magit-todos-exclude-globs))
+                 (unless magit-todos-submodule-list
+                   (--map (list "--glob" (concat "!" it))
+                          (magit-list-module-paths)))))
 
 (magit-todos-defscanner "git diff"
   ;; NOTE: This scanner implements the regexp *searching* in elisp rather than in the
@@ -1343,6 +1410,11 @@ When SYNC is non-nil, match items are returned."
                            (list "-o" "("
                                  (--map (list "-iname" it)
                                         magit-todos-exclude-globs)
+                                 ")" "-prune"))
+                         (unless magit-todos-submodule-list
+                           (list "-o" "("
+                                 (--map (list "-iname" it)
+                                        (magit-list-module-paths))
                                  ")" "-prune")))
                    (list "-o" "-type" "f")
                    ;; NOTE: This uses "grep -P", i.e. "Interpret the pattern as a
@@ -1373,6 +1445,7 @@ When SYNC is non-nil, match items are returned."
 ;; Helm or Ivy to be installed; it is only called after one of them is loaded.
 
 (declare-function helm-make-source "ext:helm-source")
+(declare-function helm "ext:helm")
 
 (with-eval-after-load 'helm
   (defvar helm-magit-todos-source
